﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.IO;
using System.IO.IsolatedStorage;
using System.Reflection;
using System.Windows.Resources;
using System.Windows.Markup;
using System.Text;

namespace SLTest
{
    public class PackageUtil
    {
        public static T GetObject<T>(Stream packageStream,
            string assemblyName,
            string typeName)
        {
            Assembly asm = LoadAssemblyFromXap(packageStream, assemblyName);
            return (T)asm.CreateInstance(typeName);
        }

        public static Stream GetResource(Stream packageStream,
            string streamName,
            string contentType)
        {
            StreamResourceInfo streamInfo =
                Application.GetResourceStream(
                   new StreamResourceInfo(packageStream, contentType),
                   new Uri(streamName, UriKind.Relative));
            return streamInfo.Stream;
        }

        public static Assembly LoadAssemblyFromXap(Stream packageStream,
            string assemblyName)
        {
            StreamResourceInfo resourceInfo = new StreamResourceInfo(packageStream, null);
            Stream resourceStream = Application.GetResourceStream(
                    resourceInfo,
                    new Uri("AppManifest.xaml", UriKind.Relative)).Stream;
            string appManifestString;
            using (StreamReader reader = new StreamReader(resourceStream))
            {
                appManifestString = reader.ReadToEnd();
                reader.Close();
            }
            Deployment deployment = (Deployment)XamlReader.Load(appManifestString);
            Assembly asm = null;
            foreach (AssemblyPart assemblyPart in deployment.Parts)
            {
                if (assemblyPart.Source.ToLower() == assemblyName.ToLower())
                {
                    string source = assemblyPart.Source;
                    //problem in sl2b2 with cache, will be fixed in sl2 final release
                    StreamResourceInfo streamInfo =
                        Application.GetResourceStream(
                           new StreamResourceInfo(packageStream, "application/binary"),
                           new Uri(source, UriKind.Relative));

                    asm = assemblyPart.Load(streamInfo.Stream);
                    break;
                }

            }
            return asm;
        }
    }

    //abstracts the WebClient 
    public class SLPackage
    {
        public enum CacheTypes
        {
            None,
            Application,
            Site
        }

        private Uri _packageUri;
        private CacheTypes _cacheType;
        private int _version;
        public class PackageEventArgs : EventArgs
        {
            private Stream _packageStream;
            private string _packageUri;
            public PackageEventArgs(Stream packageStream, string packageUri)
            {
                this._packageStream = packageStream;
                this._packageUri = packageUri;
            }
            public Stream PackageStream { get { return _packageStream; } }
            public String PackageUri { get { return _packageUri; } }
        }
        public delegate void PackageEventHandler(object sender,
                                             PackageEventArgs e);

        public event PackageEventHandler PackageDownloaded;

        public SLPackage(Uri uri, CacheTypes CacheType, int Version)
        {
            _packageUri = uri;
            _cacheType = CacheType;
            _version = Version;
        }

        public void ClearCache()
        {
            IsolatedStorageFile file = GetCacheFile();
            if (file == null)
            {
                if (file.FileExists(_packageUri.LocalPath))
                    file.DeleteFile(_packageUri.LocalPath);
                if (file.FileExists(GetInfoFile()))
                    file.DeleteFile(GetInfoFile());
            }
        }

        private string GetInfoFile()
        {
            return _packageUri.LocalPath + ".info";
        }

        public void LoadPackage()
        {
            bool needDownload;
            if (_cacheType != CacheTypes.None)
            {
                IsolatedStorageFile file = GetCacheFile();
                string localFile = _packageUri.LocalPath;
                needDownload = !file.FileExists(localFile);
                if (!needDownload)
                {
                    try
                    {
                        if (_version != -1)
                        {
                            using (IsolatedStorageFileStream stream = file.OpenFile(GetInfoFile(), FileMode.Open))
                            {                            
                                using (StreamReader reader = new StreamReader(stream))
                                {
                                    string version = reader.ReadLine();
                                    needDownload = _version > Convert.ToInt32(version);
                                }
                            }
                        }
                        if (!needDownload)
                        {
                            using (IsolatedStorageFileStream stream = file.OpenFile(localFile, FileMode.Open))
                            {
                                PackageEventArgs pe = null;
                                byte[] content = new byte[stream.Length];
                                stream.Read(content, 0, content.Length);
                                stream.Close();
                                MemoryStream mem = new MemoryStream(content);
                                pe = new PackageEventArgs(mem,
                                                 this._packageUri.OriginalString);
                                PackageDownloaded(this, pe);
                            }
                        }
                    }
                    catch (Exception)
                    {
                        needDownload = true;
                    }
                }
            }
            else
                needDownload = true;
            if (needDownload)
            {
                WebClient webClient = new WebClient();
                webClient.OpenReadCompleted += new OpenReadCompletedEventHandler(OnWebClientOpenReadCompleted);
                webClient.OpenReadAsync(_packageUri);
            }
        }

        private IsolatedStorageFile GetCacheFile()
        {
            switch (_cacheType)
            {
                case CacheTypes.Application:
                    return IsolatedStorageFile.GetUserStoreForApplication();
                case CacheTypes.Site:
                    return IsolatedStorageFile.GetUserStoreForSite();
                default:
                    return null;
            }
        }

        private void OnWebClientOpenReadCompleted(object sender,
                                   OpenReadCompletedEventArgs e)
        {
            PackageEventArgs pe = null;
            if (e.Error == null)
            {
                try
                {
                    if (_cacheType != CacheTypes.None)
                    {
                        IsolatedStorageFile file = GetCacheFile();
                        string localFile = _packageUri.LocalPath;
                        ClearCache();
                        string path = System.IO.Path.GetDirectoryName(localFile);
                        if (!file.DirectoryExists(path))
                            file.CreateDirectory(path);
                        bool readyToCache;
                        if (file.AvailableFreeSpace <= e.Result.Length)
                        {
                            long requestNewSize = file.Quota + e.Result.Length + 1024 * 1024;
                            readyToCache = file.IncreaseQuotaTo(requestNewSize);
                        }
                        else
                            readyToCache = true;
                        if (readyToCache)
                        {
                            using (IsolatedStorageFileStream writer = file.CreateFile(localFile))
                            {
                                byte[] content = new byte[e.Result.Length];
                                e.Result.Read(content, 0, content.Length);
                                writer.Write(content, 0, content.Length);
                                writer.Close();
                                using (IsolatedStorageFileStream info = file.CreateFile(GetInfoFile()))
                                {
                                    content = Encoding.UTF8.GetBytes(_version.ToString() + "\r\n");
                                    info.Write(content, 0, content.Length);
                                    content = Encoding.UTF8.GetBytes(_packageUri.OriginalString + "\r\n");
                                    info.Write(content, 0, content.Length);
                                    info.Close();
                                }
                                e.Result.Position = 0;
                            }
                        }
                    }
                    pe = new PackageEventArgs(e.Result,
                                     this._packageUri.OriginalString);
                    PackageDownloaded(this, pe);
                }
                catch
                {
                    PackageDownloaded(this, null);
                    return;
                }
            }
            else
            {
                PackageDownloaded(this, null);
                return;
            }
        }
    }
}
